from boto3.session import Session


def get_available_profiles():
    return [profile for profile in Session().available_profiles]


def get_all_regions():
    return Session().get_available_regions("dynamodb")



from textual import on
from textual.app import ComposeResult
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import  OptionList, Input

from dyno_viewer.components.query.sort_key_filter import SortKeyFilter


class KeyFilter(Widget):
    index_mode = reactive("table")

    gsi_indexes = reactive({})

    partition_key_attr_name = reactive("")
    sort_key_attr_name = reactive("")

    def compose(self) -> ComposeResult:
        yield OptionList("table", id="queryIndex")
        yield Input(placeholder="pk", id="partitionKey")
        yield SortKeyFilter(id="sortKeyFilter")

    #  on methods
    def on_mount(self):
        # select the first value for queryIndex
        option_list = self.query_one("#queryIndex")
        option_list.action_first()
        option_list.action_select()

    @on(OptionList.OptionSelected)
    def gsi_index_update(self, selected: OptionList.OptionSelected):
        self.index_mode = selected.option.prompt
        if selected.option.prompt != "table":
            new_primary_key = self.gsi_indexes[selected.option.prompt]["primaryKey"]
            new_sort_key = self.gsi_indexes[selected.option.prompt]["sortKey"]

            self.log("new_primary_key=", new_primary_key)
            self.log("new_sort_key=", new_sort_key)

            self.query_one("#partitionKey").placeholder = new_primary_key
            self.query_one("#sortKeyFilter").attr_name = new_sort_key


        else:
            self.query_one("#partitionKey").placeholder = self.partition_key_attr_name
            self.query_one("#sortKeyFilter").attr_name = self.sort_key_attr_name


    # watch methods

    def watch_gsi_indexes(self, new_gsi_indexes) -> None:
        if new_gsi_indexes:
            option_list: OptionList = self.query_one("#queryIndex")
            option_list.clear_options()
            for option in ["table", *list(new_gsi_indexes.keys())]:
                option_list.add_option(option)

            option_list.action_first()
            option_list.action_select()

    def watch_partition_key_attr_name(self, new_partition_key_attr_name) -> None:
        if new_partition_key_attr_name:
            self.query_one("#partitionKey").placeholder = new_partition_key_attr_name

    def watch_sort_key_attr_name(self, sort_key_attr_name) -> None:
        if sort_key_attr_name:
            self.query_one("#sortKeyFilter").attr_name = sort_key_attr_name



from textual import log, on, work
from textual.app import ComposeResult
from textual.containers import Vertical
from textual.events import Key, ScreenResume
from textual.message import Message
from textual.reactive import reactive
from textual.screen import Screen
from textual.widgets import Input, OptionList
from textual.worker import get_current_worker

from dyno_viewer.aws.ddb import list_all_tables


class TableSelectScreen(Screen):
    BINDINGS = [("escape", "app.pop_screen", "Pop screen")]

    dyn_client = reactive(None, always_update=True)

    next_token = reactive(None)

    def __init__(
        self, name: str | None = None, id: str | None = None, classes: str | None = None
    ) -> None:
        self.tables = []
        super().__init__(name, id, classes)

    # message classes

    class TableName(Message):
        """pass back what table was selected"""

        def __init__(self, table: str) -> None:
            self.table = table
            super().__init__()

    class TableListResult(Message):
        """return result from listing tables"""

        def __init__(self, table_result, next_token=None) -> None:
            self.table_result = table_result
            self.next_token = next_token
            super().__init__()

    def compose(self) -> ComposeResult:
        yield Input(placeholder="Search for table", id="tableSelectInput")
        yield OptionList(id="optionDdbTableList")

    # worker methods

    @work(exclusive=True, thread=True)
    def worker_list_tables(self, next_token=None):
        worker = get_current_worker()
        if not worker.is_cancelled:
            if next_token:
                list_tables_result, next_token = list_all_tables(
                    self.dyn_client,
                    Limit=10,
                    ExclusiveStartTableName=next_token,
                    paginate=False,
                )
            else:
                list_tables_result, next_token = list_all_tables(
                    self.dyn_client, Limit=10, paginate=False
                )

            self.post_message(self.TableListResult(list_tables_result, next_token))

    # on methods

    def on_mount(self) -> None:
        table_input = self.query_one(Input)
        table_input.focus()

    def on_input_submitted(self, submitted: Input.Submitted) -> None:
        if submitted.value in self.tables:
            self.post_message(self.TableName(submitted.value))
            self.app.pop_screen()

    @on(ScreenResume)
    async def paginate_tables_on_resume(self):
        if self.next_token:
            self.worker_list_tables(self.next_token)

    @on(TableListResult)
    async def on_table_list_result(self, result: TableListResult):
        new_tables = [
            table_name
            for table_name in result.table_result
            if table_name not in self.tables
        ]
        self.tables.extend(new_tables)
        self.next_token = result.next_token
        option_list = self.query_one(OptionList)
        option_list.add_options(self.tables)

    @on(Input.Changed, "#tableSelectInput")
    async def on_table_search_changed(self, changed: Input.Changed) -> None:
        table_list = self.query_one(OptionList)

        if changed.input.value and changed.input.value not in self.tables:
            matched_tables = [
                table for table in self.tables if changed.input.value in table
            ]
            table_list.clear_options()
            if matched_tables:
                table_list.add_options(matched_tables)
        elif table_list.option_count != len(self.tables):
            table_list.clear_options()
            table_list.add_options(self.tables)

    @on(OptionList.OptionSelected, "#optionDdbTableList")
    def on_table_selected(self, option_selected: OptionList.OptionSelected) -> None:
        table_input = self.query_one(Input)
        table_input.value = str(option_selected.option.prompt)
        table_input.focus()

    # watch methods

    async def watch_dyn_client(self, new_dyn_client) -> None:
        if new_dyn_client:
            option_list = self.query_one(OptionList)
            option_list.clear_options()
            self.worker_list_tables()

    async def watch_next_token(self):
        if self.is_current and self.next_token:
            self.worker_list_tables(self.next_token)



from textual.widgets import (
    DataTable,
)
from textual import log
from dyno_viewer.app_types import TableInfo


class DataDynTable(DataTable):
    def add_dyn_data(self, table_info: TableInfo, data: list[dict]) -> None:
        """
        add dynamodb table data with keys sorted to always be first
        """
        log("add dynamodb data from table")
        key_schema = table_info["keySchema"]
        gsis = table_info["gsi"]

        gsis_col = [
            key for gsi in gsis.values() for key in [gsi["primaryKey"], gsi["sortKey"]]
        ]
        log.info(f"{len(gsis_col)} gsi cols")

        cols = [key_schema["primaryKey"], key_schema["sortKey"], *gsis_col]
        data_cols = set(
            [attrKey for item in data for attrKey in item if attrKey not in cols]
        )
        log.info(f"{len(data_cols)} other cols")
        cols.extend(data_cols)

        log.info(f"{len(cols)} total cols")

        self.clear(columns=True)
        for col in cols:
            self.add_column(col, key=col)

        log.info("col keys=", [str(col.label) for col in self.columns.values()])

        rows = [[item.get(col) for col in cols] for item in data]
        log.info(f"{len(rows)} total rows")
        self.add_rows(rows)

    def add_dyn_data_existing(self, data: list[dict]) -> None:
        """
        add more data to table that has already been setup with dynamodb table data
        """
        if self.row_count == 0:
            raise Exception("there must be existing data")

        cols_not_exist = set(
            [
                attrKey
                for item in data
                for attrKey in item
                if attrKey not in self.columns
            ]
        )

        if cols_not_exist:
            log.info(f"adding cols to existing: {cols_not_exist}")
            for col in list(cols_not_exist):
                self.add_column(col, key=col)


            log.info(f"added cols to existing: {cols_not_exist}")

        rows = [[item.get(col.value) for col in self.columns.keys()] for item in data]
        log.info(f"adding rows to existing")
        self.add_rows(rows)
        log.info(f"added rows to existing")



from textual.app import ComposeResult
from textual.widget import Widget
from textual.widgets import Input, Button, RadioSet, Select, Label


class FilterQuery(Widget):
    def compose(self) -> ComposeResult:
        yield Input(placeholder="attr", id="attr")
        yield Label("Type")
        yield Select(
            [
                (line, line)
                for line in [
                    "string",
                    "number",
                    "binary",
                    "boolean",
                    "map",
                    "list",
                    "set",
                ]
            ],
            prompt="type",
            value="string",
            id="attrType",
        )

        yield Label("Condition")
        yield Select(
            [
                (line, line)
                for line in [
                    "==",
                    ">",
                    "<",
                    "<=",
                    ">=",
                    "!=",
                    "between",
                    "in",
                    "attribute_exists",
                    "attribute_not_exists",
                    "attribute_type",
                    "begins_with",
                    "contains",
                    "size",
                ]
            ],
            prompt="Condition",
            value="==",
            id="condition",
        )
        yield Input(placeholder="value", id="attrValue")
        if self.id != "sortKeyFilter":
            yield Button("remove filter", id="removeFilter")

    #  on methods

    def on_mount(self) -> None:
        for radio_set in self.query(RadioSet):
            radio_set.display = False

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if str(event.button.label) == "type":
            radio_set = self.query_one("#attrType")
            radio_set.display = False if radio_set.display else True
            self.scroll_visible()
        if str(event.button.label) == "condition":
            radio_set = self.query_one("#condition")
            radio_set.display = False if radio_set.display else True
            self.scroll_visible()
        if event.button.id == "removeFilter":
            self.remove()



from decimal import Decimal
import io
import csv

def chunk(list_to_split, amount):
    """split a list into n amount"""
    for i in range(0, len(list_to_split), amount):
        yield list_to_split[i : i + amount]


def format_output(value):
    if isinstance(value, Decimal):
        return str(value)
    else:
        return value

def output_to_csv_str(iterable):
        output = io.StringIO()
        iterable = [format_output(item) for item in iterable]
        writer = csv.writer(output)
        writer.writerow(iterable)
        return output.getvalue()


from textual.app import ComposeResult
from textual.css.query import NoMatches
from textual.widgets import ListItem, ListView, Button, Input, Switch, Label, RadioSet
from textual.widget import Widget
from textual.widgets import Footer
from textual.containers import Container, Horizontal
from textual.screen import Screen
from textual.message import Message
from textual.reactive import reactive
from textual import log, on
from dyno_viewer.components.query.filter_query import FilterQuery
from dyno_viewer.components.query.key_filter import KeyFilter
from boto3.dynamodb.conditions import Key, Attr
from dyno_viewer.aws.ddb import (
    convert_filter_exp_key_cond,
    convert_filter_exp_attr_cond,
    convert_filter_exp_value,
)
from dyno_viewer.app_types import TableInfo


class QueryScreen(Screen):
    BINDINGS = [
        ("escape", "app.pop_screen", "Pop screen"),
        (("r", "run_query", "Run Query")),
    ]

    table_info = reactive(None)
    scan_mode = reactive(False)

    class RunQuery(Message):
        """
        custom message to send back to root screen that has all query

        """

        def __init__(
            self,
            key_cond_exp: Key | None,
            filter_cond_exp: Attr | None,
            index: str | None,
        ) -> None:
            self.key_cond_exp = key_cond_exp
            self.filter_cond_exp = filter_cond_exp
            self.index = index
            super().__init__()

    def compose(self) -> ComposeResult:
        with Container(id="queryScreen"):
            yield Horizontal(
                Label("Scan "),
                Switch(name="scan", id="scanToggleSwitch"),
                id="scanToggle",
            )
            yield KeyFilter(id="keyFilter")
            yield Button("add filter", id="addFilter")
            yield Button("remove all filters", id="removeAllFilters")
            yield Footer()

    def get_primary_key(self):
        key_input = self.query_one(KeyFilter)
        primary_key_name = (
            key_input.partition_key_attr_name
            if key_input.index_mode == "table"
            else key_input.gsi_indexes[key_input.index_mode]["primaryKey"]
        )
        primary_key_value = key_input.query_one("#partitionKey").value
        log("attr primary key name=", primary_key_name)
        log("attr primary key value=", primary_key_value)
        return primary_key_name, primary_key_value

    def get_sort_key_(self):
        key_input = self.query_one(KeyFilter)
        sort_key = key_input.query_one("#sortKeyFilter")
        sort_key_name = (
            key_input.sort_key_attr_name
            if key_input.index_mode == "table"
            else key_input.gsi_indexes[key_input.index_mode]["sortKey"]
        )
        sort_key_value = sort_key.query_one("#attrValue").value
        log("attr sort key value=", sort_key_value)
        cond = sort_key.query_one("#condition").value
        return sort_key_name, sort_key_value, cond

    def get_key_query(self) -> Key | None:
        log("generating key expression from input data")
        primary_key_name, primary_key_value = self.get_primary_key()

        if not primary_key_value:
            return

        sort_key_name, sort_key_value, cond = self.get_sort_key_()

        if sort_key_value:
            return Key(primary_key_name).eq(
                primary_key_value
            ) & convert_filter_exp_key_cond(cond, sort_key_name, sort_key_value)
        return Key(primary_key_name).eq(primary_key_value)

    def get_filter_queries(self) -> Attr | None:
        exp = None

        for filter in self.query(FilterQuery):
            attr_name = filter.query_one("#attr").value
            attr_type = getattr(filter.query_one("#attrType"), "value", "")

            attr_value = str(getattr(filter.query_one("#attrValue"), "value", ""))
            cond = getattr(filter.query_one("#condition"), "value", "")

            if exp:
                exp = exp & convert_filter_exp_attr_cond(
                    cond, attr_name, convert_filter_exp_value(attr_value, attr_type)
                )
            else:
                exp = convert_filter_exp_attr_cond(
                    cond, attr_name, convert_filter_exp_value(attr_value, attr_type)
                )
        return exp

    def update_key_schema(self, key_query):
        key_query.partition_key_attr_name = self.table_info["keySchema"]["primaryKey"]
        key_query.sort_key_attr_name = self.table_info["keySchema"]["sortKey"]
        key_query.gsi_indexes = self.table_info["gsi"]

    # action methods

    def action_run_query(self) -> None:
        key_filter = self.query("#keyFilter")

        if key_filter:
            key_cond_exp = self.get_key_query()
            index_mode = self.query_one(KeyFilter).index_mode
        else:
            key_cond_exp = None
            index_mode = "table"

        filter_cond_exp = self.get_filter_queries()

        if key_cond_exp or filter_cond_exp or self.scan_mode:
            self.post_message(
                self.RunQuery(
                    key_cond_exp,
                    filter_cond_exp,
                    index_mode,
                )
            )
            self.app.pop_screen()

    # on methods:
    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "addFilter":
            self.query_one("#queryScreen").mount(FilterQuery())
            self.scroll_visible()
        elif event.button.id == "removeAllFilters":
            for filter in self.query(FilterQuery):
                if filter.id != "sortKeyFilter":
                    filter.remove()
            self.scroll_visible()

    def on_mount(self):
        if self.table_info:
            key_query = self.query_one(KeyFilter)
            self.update_key_schema(key_query)

    @on(Switch.Changed, "#scanToggleSwitch")
    def toggle_scan_mode(self, changed: Switch.Changed) -> None:
        self.scan_mode = changed.value
        if changed.value:
            key_filter = self.query_one(KeyFilter)
            key_filter.remove()
        else:
            self.mount(KeyFilter(id="keyFilter"), after="#scanToggle")

    # watcher methods

    def watch_table_info(self, new_table_info: TableInfo) -> None:
        if new_table_info:
            try:
                key_query = self.query_one(KeyFilter)
                self.update_key_schema(key_query)
            except NoMatches:
                return
            except Exception:
                raise



from textual.app import ComposeResult
from textual.widgets import (
    Label,
)
from textual.screen import Screen
from textual.containers import Vertical
from textual.widgets import Button
from textual import log
from botocore.exceptions import ClientError


class ErrorScreen(Screen):
    BINDINGS = [("escape", "app.pop_screen", "Pop screen")]

    def __init__(
        self,
        error_msg: str,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ) -> None:
        self.error_msg = error_msg
        super().__init__(name, id, classes)

    def compose(self) -> ComposeResult:
        yield Vertical(Label(self.error_msg), Button("OK"))

    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.app.pop_screen()



import logging

import simplejson as json
from decimal import Decimal as D

from boto3.dynamodb.conditions import Attr, Key
from dynamodb_json import json_util as dyn_json
from boto3.session import Session
import boto3
from botocore.exceptions import ClientError
from decimal import Decimal
import re
from textual import log


LOG_LEVEL = logging.INFO


def get_logger():
    import logging

    logging.basicConfig()
    logger = logging.getLogger()
    logger.setLevel(LOG_LEVEL)
    return logger


logger = get_logger()


def table_client_exist(table_name, region_name, profile_name):
    try:
        client = get_table(table_name, region_name, profile_name)
        if client.table_status in ("CREATING", "UPDATING", "ACTIVE"):
            return client
    except Exception as error:
        if error.response["Error"]["Code"] in [
            "ResourceNotFoundException",
        ]:
            return
        raise error


def get_table(table_name, region_name, profile_name):
    return get_dyn_resource(region_name, profile_name).Table(table_name)


def get_dyn_resource(region_name, profile_name):
    return (
        Session(profile_name=profile_name, region_name=region_name).resource("dynamodb")
        if profile_name
        else boto3.resource("dynamodb", region_name=region_name)
    )


def get_table_client(table, region_name="ap-southeast-2", profile_name=None):
    return (
        get_table(table, region_name, profile_name) if isinstance(table, str) else table
    )


def get_ddb_client(region_name="ap-southeast-2", profile_name=None):
    return (
        Session(profile_name=profile_name, region_name=region_name).client("dynamodb")
        if profile_name
        else boto3.client("dynamodb", region_name=region_name)
    )


def list_all_tables(client=None, paginate=True, **kwargs):
    ddb_client = client or get_ddb_client(client)
    tables = []
    result = ddb_client.list_tables(**kwargs)
    tables.extend(result["TableNames"])

    if paginate:
        while "LastEvaluatedTableName" in result:
            tables.extend(result["TableNames"])
            result = ddb_client.list_tables(
                **kwargs, ExclusiveStartTableName=result["LastEvaluatedTableName"]
            )

        return tables
    return tables, result.get("LastEvaluatedTableName")


def get_item(table, item_key, return_none=False, consistent_read=False):
    """
    Get an item using the item key.

    :param table: name or client of the dynamodb table
    :param item_key: e.g {"pk":"partition_key", "sk":"sort_key"}
    :param return_none: return None if the item doesn't exist.
    :return:
    """
    resp = get_table_client(table).get_item(
        Key=item_key, ConsistentRead=consistent_read
    )
    if not return_none:
        return resp["Item"]
    else:
        return resp.get("Item")


def get_items(table_name, item_keys, return_none=False, **kwargs):
    """
    Use the BatchGetItem API to bulk read items.
    """

    def make_key(key):
        return {"pk": {"S": key["pk"]}, "sk": {"S": key["sk"]}}

    items = []
    keys = [make_key(d) for d in item_keys]
    logger.info(f"Batch reading {len(keys)} item keys.")
    batch_size = 100  # the max ddb supports.
    batches = [keys[n : n + batch_size] for n in range(0, len(keys), batch_size)]

    client = get_ddb_client()
    request = {"RequestItems": {table_name: {"Keys": None, **kwargs}}}

    for n, batch in enumerate(batches):
        request["RequestItems"][table_name]["Keys"] = batch
        res = client.batch_get_item(**request)
        for k, v in res["Responses"].items():
            items.extend(v)
        logger.info(f"{n}/{len(batches)}")

    return items


def query_items(
    table,
    paginate=True,
    **query_kwargs,
):
    items = []
    table_client = get_table_client(table)
    resp = table_client.query(**query_kwargs)
    items.extend(resp["Items"])
    if paginate:
        while "LastEvaluatedKey" in resp:
            resp = table_client.query(
                **query_kwargs, ExclusiveStartKey=resp["LastEvaluatedKey"]
            )
            items.extend(resp["Items"])
        return items

    return items, resp.get("LastEvaluatedKey")


def query_first(table, return_none=False, **query_kwargs):
    items = query_items(table, **query_kwargs)
    items_count = len(items)

    if items_count == 0 and return_none:
        return None

    if items_count == 0:
        raise Exception(f"Item not found")

    return items[0]


def check_attr_exists(table, pk, attr, sk=None):
    """
    Check an attribute exists given a pk/sk.
    """
    return bool(
        len(
            get_table_client(table).query(
                KeyConditionExpression=Key("pk").eq(pk) & Key("sk").eq(sk),
                FilterExpression=Attr(attr).exists(),
            )["Items"]
        )
    )


def scan_items(table, paginate=True, **query_kwargs):
    """Debugging utility."""
    items = []
    table_client = get_table_client(table)
    resp = table_client.scan(**query_kwargs)
    items.extend(resp["Items"])
    if paginate:
        while "LastEvaluatedKey" in resp:
            resp = table_client.scan(
                **query_kwargs, ExclusiveStartKey=resp["LastEvaluatedKey"]
            )
            items.extend(resp["Items"])
        return items

    return items, resp.get("LastEvaluatedKey")


def covert_comparator_exp(cond, attr_name, value, is_key=True) -> Key | Attr:
    attr_class = Key if is_key else Attr
    if cond == "==":
        return attr_class(attr_name).eq(value)
    if cond == "!=":
        return attr_class(attr_name).ne(value)
    elif cond == ">=":
        return attr_class(attr_name).gte(value)
    elif cond == ">":
        return attr_class(attr_name).gt(value)
    elif cond == "<=":
        return attr_class(attr_name).lte(value)
    elif cond == "<":
        return attr_class(attr_name).lt(value)


def convert_filter_exp_key_cond(cond, attr_name, value) -> Key:
    comparator = covert_comparator_exp(cond, attr_name, value)

    if comparator:
        return comparator

    if cond == "between":
        return Key(attr_name).between(value[0], value[1])
    if cond == "begins_with":
        return Key(attr_name).begins_with(value)

    raise Exception("passed incorrect condition for key filter expression")


def convert_filter_exp_attr_cond(cond, attr_name, value=None) -> Attr:
    comparator = covert_comparator_exp(cond, attr_name, value, is_key=False)

    if comparator:
        return comparator

    if cond == "between":
        return Attr(attr_name).between(value[0], value[1])
    if cond == "begins_with":
        return Attr(attr_name).begins_with(value)
    if cond == "in":
        return Attr(attr_name).is_in(value)
    if cond == "attribute_exists":
        return Attr(attr_name).exists()
    if cond == "attribute_not_exists":
        return Attr(attr_name).not_exists()
    if cond == "attribute_type":
        return Attr(attr_name).attribute_type(value)
    if cond == "contains":
        return Attr(attr_name).contains(value)
    if cond == "size":
        return Attr(attr_name).size()

    raise Exception("passed incorrect condition for key filter expression")


def convert_filter_exp_value(value: str, type: str):
    if type == "number":
        return Decimal(value)
    elif type == "list":
        return list(value)
    elif type == "map":
        return json.loads(value)
    elif type == "boolean":
        return bool(value)
    elif type == "set":
        value_formatted = re.sub(r"[\(\)]", "", value)
        return set(value_formatted.split(","))
    else:
        return value


def float_to_decimal(payload):
    return json.loads(json.dumps(payload), parse_float=D)


def serialise_dynamodb_json(json_obj):
    return json.loads(dyn_json.dumps(json_obj))



from textual.app import ComposeResult
from textual.widgets import (
    ListItem,
    ListView,
    Label,
)
from textual.screen import Screen
from textual.message import Message
from textual import log

from dyno_viewer.aws.session import get_available_profiles


class ProfileSelectScreen(Screen):
    BINDINGS = [("escape", "app.pop_screen", "Pop screen")]

    class ProfileSelected(Message):
        def __init__(self, profile: str) -> None:
            self.profile = profile
            super().__init__()

    def compose(self) -> ComposeResult:
        yield ListView(
            *[
                ListItem(Label(profile), id=profile)
                for profile in get_available_profiles()
            ],
            id="profiles",
        )

    async def on_list_view_selected(self, selected) -> None:
        self.post_message(self.ProfileSelected(selected.item.id))
        self.app.pop_screen()



from textual.app import App, ComposeResult
from textual.widgets import (
    Footer,
)
from textual.reactive import reactive
from dyno_viewer.aws.session import get_available_profiles
from dyno_viewer.aws.ddb import (
    get_ddb_client,
    query_items,
    scan_items,
    table_client_exist,
)
from dyno_viewer.components.screens import (
    ProfileSelectScreen,
    RegionSelectScreen,
    TableSelectScreen,
    QueryScreen,
    HelpMenu,
)
from dyno_viewer.components.table import DataDynTable
from textual.worker import get_current_worker
from textual.binding import Binding
from textual import work, log, on
import pyclip
from itertools import cycle
from dyno_viewer.app_types import TableInfo
from textual.message import Message
from dyno_viewer.util import output_to_csv_str
from dyno_viewer.util.util import format_output

cursors = cycle(["column", "row", "cell"])


class UpdateDynDataTable(Message):
    def __init__(self, data, next_token, update_existing_data=False) -> None:
        self.data = data
        self.next_token = next_token
        self.update_existing_data = update_existing_data
        super().__init__()


class UpdateDynTableInfo(Message):
    def __init__(self, table_info: TableInfo) -> None:
        self.table_info = table_info
        super().__init__()


class DynCli(App):
    BINDINGS = [
        ("x", "exit", "Exit"),
        ("p", "push_screen('profile')", "Profile"),
        ("t", "push_screen('tableSelect')", "Table"),
        ("r", "push_screen('regionSelect')", "Region"),
        ("q", "push_screen_query", "Query"),
        ("?", "push_screen('help')", "help"),
        Binding("ctrl+c", "copy_table_data", "Copy", show=False),
        Binding("ctrl+r", "change_cursor_type", "Change Cursor type", show=False),
    ]
    SCREENS = {
        "tableSelect": TableSelectScreen(),
        "regionSelect": RegionSelectScreen(),
        "profile": ProfileSelectScreen(),
        "query": QueryScreen(),
        "help": HelpMenu(),
    }

    CSS_PATH = ["components/css/query.tcss", "components/css/table.tcss"]

    profiles = reactive(get_available_profiles())

    aws_profile = reactive(None)

    table_name = reactive("")

    aws_region = reactive("ap-southeast-2")

    dyn_query_params = reactive({})

    # set always_update=True because otherwise textual thinks that the client hasn't changed when it actually has :(
    table_client = reactive(None, always_update=True)

    dyn_client = reactive(get_ddb_client())

    table_info = reactive(None)

    def compose(self) -> ComposeResult:
        yield DataDynTable()
        yield Footer()

    def update_table_client(self):
        if self.table_name != "":
            log.info(f"updating table client with profile {self.aws_profile}")
            new_table_client = table_client_exist(
                self.table_name, self.aws_region, self.aws_profile
            )
            if new_table_client:
                self.table_client = new_table_client
            else:
                table = self.query_one(DataDynTable)
                table.clear()

    def set_pagination_token(self, next_token: str | None) -> None:
        if next_token:
            self.dyn_query_params["ExclusiveStartKey"] = next_token
        else:
            self.dyn_query_params.pop("ExclusiveStartKey", None)

    # worker methods

    @work(exclusive=True, group="update_dyn_table_info", thread=True)
    def get_dyn_table_info(self) -> None:
        worker = get_current_worker()
        if not worker.is_cancelled:
            # temp disable logging doesn't work
            self.log(f"updating table info")
            self.log("key schema=", self.table_client.key_schema)
            self.log("gsi schema=", self.table_client.global_secondary_indexes)
            main_keys = {
                ("primaryKey" if key["KeyType"] == "HASH" else "sortKey"): key[
                    "AttributeName"
                ]
                for key in self.table_client.key_schema
            }

            gsi_keys = {
                gsi["IndexName"]: {
                    ("primaryKey" if key["KeyType"] == "HASH" else "sortKey"): key[
                        "AttributeName"
                    ]
                    for key in gsi["KeySchema"]
                }
                for gsi in self.table_client.global_secondary_indexes or []
            }

            self.post_message(
                UpdateDynTableInfo({"keySchema": main_keys, "gsi": gsi_keys})
            )

    @work(exclusive=True, group="dyn_table_query", thread=True)
    def run_table_query(self, dyn_query_params, update_existing=False):
        worker = get_current_worker()
        if not worker.is_cancelled:
            self.log("dyn_params=", dyn_query_params)
            result, next_token = (
                query_items(
                    self.table_client,
                    paginate=False,
                    Limit=50,
                    **dyn_query_params,
                )
                if "KeyConditionExpression" in dyn_query_params
                else scan_items(
                    self.table_client,
                    paginate=False,
                    Limit=50,
                    **dyn_query_params,
                )
            )

            self.post_message(UpdateDynDataTable(result, next_token, update_existing))

    # on methods

    def on_mount(self):
        table = self.query_one(DataDynTable)
        table.focus()

    @on(DataDynTable.CellHighlighted)
    async def paginate_dyn_data(
        self, highlighted: DataDynTable.CellHighlighted
    ) -> None:
        if highlighted.coordinate.row == highlighted.data_table.row_count - 1:
            if "ExclusiveStartKey" in self.dyn_query_params:
                log.info("adding more items")

                self.run_table_query(self.dyn_query_params, update_existing=True)

    @on(UpdateDynTableInfo)
    async def update_table_info(self, update: UpdateDynTableInfo) -> None:
        self.table_info = update.table_info

    async def on_region_select_screen_region_selected(
        self, selected_region: RegionSelectScreen.RegionSelected
    ) -> None:
        self.aws_region = selected_region.region
        self.dyn_client = get_ddb_client(selected_region.region, self.aws_profile)
        self.update_table_client()

    async def on_table_select_screen_table_name(
        self,
        new_table_name: TableSelectScreen.TableName,
    ) -> None:
        if self.table_name != new_table_name:
            self.table_name = new_table_name.table
            self.update_table_client()

    async def on_profile_select_screen_profile_selected(
        self, selected_profile: ProfileSelectScreen.ProfileSelected
    ) -> None:
        self.aws_profile = selected_profile.profile
        log.info(f"{self.aws_profile} profile selected")
        self.dyn_client = get_ddb_client(
            region_name=self.aws_region, profile_name=self.aws_profile
        )
        self.update_table_client()

    async def on_query_screen_run_query(self, run_query: QueryScreen.RunQuery) -> None:
        params = (
            {"KeyConditionExpression": run_query.key_cond_exp}
            if run_query.key_cond_exp
            else {}
        )

        if run_query.filter_cond_exp:
            params["FilterExpression"] = run_query.filter_cond_exp
        
        if run_query.index != "table":
            params["IndexName"] = run_query.index
        
        self.dyn_query_params = params
        self.run_table_query(params)

    async def on_update_dyn_data_table(self, update_data: UpdateDynDataTable) -> None:
        table = self.query_one(DataDynTable)
        if update_data.update_existing_data:
            table.add_dyn_data_existing(update_data.data)
            self.set_pagination_token(update_data.next_token)
        else:
            table.add_dyn_data(self.table_info, update_data.data)
            self.set_pagination_token(update_data.next_token)

    # action methods

    async def action_exit(self) -> None:
        table = self.query_one(DataDynTable)
        # ensure we don't have any dirty data for next time app runs
        table.clear()
        self.app.exit()

    async def action_push_screen_query(self) -> None:
        if self.table_client:
            self.push_screen("query")
        else:
            self.notify("No table selected")

    async def action_copy_table_data(self) -> None:
        query_table = self.query(DataDynTable)
        if query_table:
            table = query_table[0]
            if table.row_count > 0:
                if table.cursor_type == "cell":
                    cell = table.get_cell_at(table.cursor_coordinate)
                    if cell is not None:
                        pyclip.copy(format_output(cell))
                elif table.cursor_type == "row":
                    row = table.get_row_at(table.cursor_row)
                    if row:
                        pyclip.copy(output_to_csv_str(row))
                elif table.cursor_type == "column":
                    col = table.get_column_at(table.cursor_column)
                    if col:
                        pyclip.copy(output_to_csv_str(col))

    async def action_change_cursor_type(self) -> None:
        query_table = self.query(DataDynTable)
        if query_table:
            table = query_table[0]
            next_cursor = next(cursors)
            self.notify(f"selection mode: {next_cursor}", timeout=1)
            table.cursor_type = next_cursor

    # watcher methods

    async def watch_table_client(self, new_table_client) -> None:
        """update DynTable with new table data"""
        if new_table_client:
            log.info("table client changed and table found, Update table data")
            self.get_dyn_table_info()
            self.run_table_query(self.dyn_query_params)
        else:
            log.info("table client changed and table not found, Clear table data")
            self.query_one(DataDynTable).clear()

    def watch_dyn_client(self, new_dyn_client):
        with self.SCREENS["tableSelect"].prevent(TableSelectScreen.TableName):
            self.SCREENS["tableSelect"].dyn_client = new_dyn_client

    def watch_table_info(self, new_table_info: TableInfo) -> None:
        with self.SCREENS["query"].prevent(QueryScreen.RunQuery):
            self.SCREENS["query"].table_info = new_table_info


def run() -> None:
    app = DynCli()
    app.run()



from textual.app import ComposeResult
from textual.widgets import (
    ListItem,
    ListView,
    Label,
)
from textual.screen import Screen
from textual.message import Message
from dyno_viewer.aws.session import get_all_regions
from textual import log


class RegionSelectScreen(Screen):
    BINDINGS = [("escape", "app.pop_screen", "Pop screen")]

    class RegionSelected(Message):
        """"""

        def __init__(self, region: str) -> None:
            self.region = region
            super().__init__()

    def compose(self) -> ComposeResult:
        yield ListView(
            *[ListItem(Label(region), id=region) for region in get_all_regions()],
            id="regions",
        )

    async def on_list_view_selected(self, selected) -> None:
        self.post_message(self.RegionSelected(selected.item.id))
        self.app.pop_screen()



from textual.app import ComposeResult
from textual.widgets import (
    Label,
)
from textual.screen import Screen
from textual.containers import Vertical
from textual.widgets import Button, Markdown
import os



class HelpMenu(Screen):
    BINDINGS = [("escape", "app.pop_screen", "Pop screen")]

    def compose(self) -> ComposeResult:
        parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
        with open(os.path.join(parent_dir, "help","help-doc.md")) as f:
            text = f.read()
            yield Markdown(text, id="helpMarkdown")


